from collections.abc import Callable
from typing import get_type_hints

import pydantic

from liblaf.cherries import pathutils as _path
from liblaf.cherries import plugin, presets


def run[C: pydantic.BaseModel, T](main: Callable[[C], T]) -> T:
    run: plugin.Run = start()
    type_hints: dict[str, type[C]] = get_type_hints(main)
    cls: type[C] = next(iter(type_hints.values()))
    cfg: C = cls()
    run.log_param("cherries.config", cfg.model_dump(mode="json"))
    try:
        ret: T = main(cfg)
    except BaseException as e:
        if isinstance(e, KeyboardInterrupt):
            run.end("KILLED")
            raise
        run.end("FAILED")
        raise
    else:
        run.end()
        return ret


def start() -> plugin.Run:
    run: plugin.Run = presets.default()
    run.start()
    run.log_src(_path.entrypoint(absolute=True))
    run.set_tag("cherries.entrypoint", _path.entrypoint(absolute=False))
    run.set_tag("cherries.exp-dir", _path.exp_dir(absolute=False))
    return plugin.run


def end() -> None:
    plugin.run.end()



import datetime
import functools
from pathlib import Path

import attrs
import mlflow

from liblaf.cherries import pathutils as _path
from liblaf.cherries.typed import PathLike

from ._abc import End, LogArtifact, LogArtifacts, LogMetric, LogParam, SetTag, Start


@attrs.define
class Run:
    end: End = attrs.field(factory=End, init=False)
    log_artifact: LogArtifact = attrs.field(factory=LogArtifact, init=False)
    log_artifacts: LogArtifacts = attrs.field(factory=LogArtifacts, init=False)
    log_metric: LogMetric = attrs.field(factory=LogMetric, init=False)
    log_param: LogParam = attrs.field(factory=LogParam, init=False)
    set_tag: SetTag = attrs.field(factory=SetTag, init=False)
    start: Start = attrs.field(factory=Start, init=False)

    @property
    def active_run(self) -> mlflow.ActiveRun:
        return mlflow.active_run()  # pyright: ignore[reportReturnType]

    @functools.cached_property
    def exp_dir(self) -> Path:
        return _path.exp_dir(absolute=True)

    @property
    def exp_id(self) -> str:
        return self.active_run.info.experiment_id

    @property
    def exp_name(self) -> str:
        return self.exp_id

    @property
    def exp_url(self) -> str:
        tracking_uri: str = self.tracking_uri.rstrip("/")
        return f"{tracking_uri}/#/experiments/{self.exp_id}"

    @property
    def tracking_uri(self) -> str:
        return mlflow.get_tracking_uri()

    @property
    def run_id(self) -> str:
        return self.active_run.info.run_id

    @property
    def run_name(self) -> str:
        return self.active_run.info.run_name  # pyright: ignore[reportReturnType]

    @property
    def run_url(self) -> str:
        return f"{self.exp_url}/runs/{self.run_id}"

    @property
    def start_time(self) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(
            self.active_run.info.start_time / 1000, tz=datetime.UTC
        ).astimezone()

    def log_input(
        self, local_path: PathLike, artifact_path: PathLike | None = "inputs", **kwargs
    ) -> Path:
        return self.log_artifact(local_path, artifact_path, **kwargs)

    def log_inputs(
        self, local_dir: PathLike, artifact_path: PathLike | None = "inputs", **kwargs
    ) -> Path:
        return self.log_artifacts(local_dir, artifact_path, **kwargs)

    def log_output(
        self, local_path: PathLike, artifact_path: PathLike | None = "outputs", **kwargs
    ) -> Path:
        return self.log_artifact(local_path, artifact_path, **kwargs)

    def log_outputs(
        self, local_dir: PathLike, artifact_path: PathLike | None = "outputs", **kwargs
    ) -> Path:
        return self.log_artifacts(local_dir, artifact_path, **kwargs)

    def log_src(
        self, local_path: PathLike, artifact_path: PathLike | None = "src", **kwargs
    ) -> Path:
        return self.log_artifact(local_path, artifact_path, **kwargs)


run = Run()
end: End = run.end
log_artifact: LogArtifact = run.log_artifact
log_artifacts: LogArtifacts = run.log_artifacts
log_metric: LogMetric = run.log_metric
log_param: LogParam = run.log_param
set_tag: SetTag = run.set_tag
start: Start = run.start
log_input = run.log_input
log_inputs = run.log_inputs
log_output = run.log_output
log_outputs = run.log_outputs
log_src = run.log_src



from ._convert import as_os_path, as_path, as_posix
from ._path import entrypoint, exp_dir, git_root, git_root_safe
from ._special import config, data, inputs, outputs, params, path, src

__all__ = [
    "as_os_path",
    "as_path",
    "as_posix",
    "config",
    "data",
    "entrypoint",
    "exp_dir",
    "git_root",
    "git_root_safe",
    "inputs",
    "outputs",
    "params",
    "path",
    "src",
]



import sys
from pathlib import Path

import git
import git.exc
from loguru import logger

from liblaf.cherries import utils


@utils.cache
def entrypoint(*, absolute: bool = False) -> Path:
    if absolute:
        return _entrypoint_absolute()
    return _entrypoint_relative()


@utils.cache
def exp_dir(*, absolute: bool = False) -> Path:
    if absolute:
        return _exp_dir_absolute()
    return _exp_dir_relative()


@utils.cache
def git_root() -> Path:
    entrypoint: Path = _entrypoint_absolute()
    repo = git.Repo(entrypoint, search_parent_directories=True)
    return Path(repo.working_dir)


@utils.cache
def git_root_safe() -> Path:
    try:
        return git_root()
    except git.exc.InvalidGitRepositoryError:
        logger.warning("Not in a git repository, using current directory")
        return _entrypoint_absolute().parent


@utils.cache
def _entrypoint_absolute() -> Path:
    path = Path(sys.argv[0])
    return path.absolute()


@utils.cache
def _entrypoint_relative() -> Path:
    path: Path = _entrypoint_absolute()
    return path.relative_to(git_root_safe())


@utils.cache
def _exp_dir_absolute() -> Path:
    entrypoint: Path = _entrypoint_absolute()
    for path in entrypoint.parents:
        if (path / "exp.cherries.toml").is_file():
            return path
        if (path / "src").is_dir():
            return path
    return git_root_safe()


@utils.cache
def _exp_dir_relative() -> Path:
    absolute: Path = _exp_dir_absolute()
    root: Path = git_root_safe()
    return absolute.relative_to(root)



from . import info, pathutils, plugin, presets, utils
from ._config import BaseConfig
from ._run import end, run, start
from .pathutils import (
    as_os_path,
    as_path,
    as_posix,
    config,
    data,
    entrypoint,
    exp_dir,
    git_root,
    git_root_safe,
    inputs,
    outputs,
    params,
    path,
    src,
)
from .plugin import (
    End,
    LogArtifact,
    LogArtifacts,
    LoggingEnd,
    LoggingStart,
    LogMetric,
    LogParam,
    Plugin,
    Run,
    RunStatus,
    Start,
    log_artifact,
    log_artifacts,
    log_input,
    log_inputs,
    log_metric,
    log_output,
    log_outputs,
    log_param,
    log_src,
    set_tag,
)

__all__ = [
    "BaseConfig",
    "End",
    "LogArtifact",
    "LogArtifacts",
    "LogMetric",
    "LogParam",
    "LoggingEnd",
    "LoggingStart",
    "Plugin",
    "Run",
    "RunStatus",
    "Start",
    "as_os_path",
    "as_path",
    "as_posix",
    "config",
    "data",
    "end",
    "entrypoint",
    "exp_dir",
    "git_root",
    "git_root_safe",
    "info",
    "inputs",
    "log_artifact",
    "log_artifacts",
    "log_input",
    "log_inputs",
    "log_metric",
    "log_output",
    "log_outputs",
    "log_param",
    "log_src",
    "outputs",
    "params",
    "path",
    "pathutils",
    "plugin",
    "presets",
    "run",
    "set_tag",
    "src",
    "start",
    "utils",
]



import pydantic_settings as ps


class BaseConfig(ps.BaseSettings):
    model_config = ps.SettingsConfigDict(cli_parse_args=True, yaml_file="params.yaml")

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[ps.BaseSettings],
        init_settings: ps.PydanticBaseSettingsSource,
        env_settings: ps.PydanticBaseSettingsSource,
        dotenv_settings: ps.PydanticBaseSettingsSource,
        file_secret_settings: ps.PydanticBaseSettingsSource,
    ) -> tuple[ps.PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            env_settings,
            dotenv_settings,
            file_secret_settings,
            ps.YamlConfigSettingsSource(settings_cls),
        )



from liblaf.cherries import plugin


def default() -> plugin.Run:
    plugin.run.end.add(
        plugin.LoggingEnd(),
        plugin.GitEnd(),
        plugin.DvcEnd(),
        plugin.MlflowEnd(),
    )
    plugin.run.log_artifact.add(
        plugin.DvcLogArtifact(),
        plugin.MlflowLogArtifact(),
    )
    plugin.run.log_artifacts.add(
        plugin.DvcLogArtifacts(),
        plugin.MlflowLogArtifacts(),
    )
    plugin.run.log_metric.add(
        plugin.MlflowLogMetric(),
    )
    plugin.run.log_param.add(
        plugin.MlflowLogParam(),
    )
    plugin.run.set_tag.add(
        plugin.MlflowSetTag(),
    )
    plugin.run.start.add(
        plugin.LoggingStart(),
        plugin.MlflowStart(),
    )
    return plugin.run



from pathlib import Path
from typing import overload

from liblaf.cherries.typed import PathLike


@overload
def as_os_path(path: PathLike) -> str: ...
@overload
def as_os_path(path: None) -> None: ...
def as_os_path(path: PathLike | None) -> str | None:
    if path is None:
        return None
    if isinstance(path, str):
        return path
    return str(path)


@overload
def as_path(path: PathLike) -> Path: ...
@overload
def as_path(path: None) -> None: ...


def as_path(path: PathLike | None) -> Path | None:
    if path is None:
        return None
    return Path(path)


@overload
def as_posix(path: PathLike) -> str: ...
@overload
def as_posix(path: None) -> None: ...
def as_posix(path: PathLike | None) -> str | None:
    if path is None:
        return None
    if isinstance(path, str):
        return path
    return Path(path).as_posix()



import subprocess as sp
from pathlib import Path
from typing import override

import attrs

from liblaf.cherries import pathutils as _path
from liblaf.cherries.typed import PathLike

from ._abc import End, LogArtifact, LogArtifacts, RunStatus


@attrs.define
class DvcEnd(End):
    @override
    def __call__(self, status: RunStatus = RunStatus.FINISHED) -> None:
        sp.run(["dvc", "status"], check=True)
        sp.run(["dvc", "push"], check=True)


@attrs.define
class DvcLogArtifact(LogArtifact):
    @override
    def __call__(
        self, local_path: PathLike, artifact_path: PathLike | None = None, **kwargs
    ) -> Path:
        local_path: Path = _path.as_path(local_path)
        sp.run(["dvc", "add", local_path], check=False)
        return local_path


@attrs.define
class DvcLogArtifacts(LogArtifacts):
    @override
    def __call__(
        self, local_dir: PathLike, artifact_path: PathLike | None = None, **kwargs
    ) -> Path:
        local_dir: Path = _path.as_path(local_dir)
        sp.run(["dvc", "add", local_dir], check=False)
        return local_dir


def check_ignore(local_path: PathLike) -> bool:
    proc: sp.CompletedProcess[bytes] = sp.run(
        ["dvc", "check-ignore", local_path], check=False
    )
    return proc.returncode == 0



from __future__ import annotations

import bisect
import enum
import functools
import operator
from collections.abc import Iterable
from pathlib import Path
from typing import Any, override

import attrs
import mlflow
import mlflow.entities
from loguru import logger

from liblaf.cherries import pathutils as _path
from liblaf.cherries.typed import PathLike


class RunStatus(enum.StrEnum):
    FAILED = mlflow.entities.RunStatus.to_string(mlflow.entities.RunStatus.FAILED)
    FINISHED = mlflow.entities.RunStatus.to_string(mlflow.entities.RunStatus.FINISHED)
    KILLED = mlflow.entities.RunStatus.to_string(mlflow.entities.RunStatus.KILLED)


@functools.total_ordering
@attrs.define
class Plugin[**P, T]:
    priority: int = attrs.field(default=0, kw_only=True, eq=True, order=True)
    _children: list[Plugin] = attrs.field(
        factory=list, eq=False, order=False, alias="children"
    )

    def __attrs_post_init__(self) -> None:
        self._children.sort(key=operator.attrgetter("priority"))

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        ret: T | None = None
        for child in self._children:
            try:
                ret = child(*args, **kwargs)
            except BaseException as err:
                if isinstance(err, (KeyboardInterrupt, SystemExit)):
                    raise
                logger.exception(child)
        return ret  # pyright: ignore[reportReturnType]

    def __lt__(self, other: Plugin) -> bool:
        if not isinstance(other, Plugin):
            return NotImplemented
        return self.priority < other.priority

    def __eq__(self, other: Plugin) -> bool:  # pyright: ignore[reportIncompatibleMethodOverride]
        if not isinstance(other, Plugin):
            return NotImplemented
        return self.priority == other.priority

    @property
    def children(self) -> list[Plugin]:
        return self._children

    def add(self, *child: Plugin) -> None:
        for c in child:
            bisect.insort(self._children, c, key=operator.attrgetter("priority"))

    def extend(self, children: Iterable[Plugin]) -> None:
        self.add(*children)

    def remove(self, child: Plugin) -> None:
        self._children.remove(child)


@attrs.define
class End(Plugin):
    @override
    def __call__(self, status: RunStatus = RunStatus.FINISHED) -> None:
        return super().__call__(status)


@attrs.define
class LogArtifact(Plugin):
    @override
    def __call__(
        self, local_path: PathLike, artifact_path: PathLike | None = None, **kwargs
    ) -> Path:
        ret: Path | None = super().__call__(local_path, artifact_path, **kwargs)
        if ret is None:
            ret = _path.as_path(local_path)
        return ret


@attrs.define
class LogArtifacts(Plugin):
    @override
    def __call__(
        self, local_dir: PathLike, artifact_path: PathLike | None = None, **kwargs
    ) -> Path:
        ret: Path | None = super().__call__(local_dir, artifact_path, **kwargs)
        if ret is None:
            ret = _path.as_path(local_dir)
        return ret


@attrs.define
class LogMetric(Plugin):
    @override
    def __call__(
        self, key: str, value: float, step: int | None = None, **kwargs
    ) -> None:
        return super().__call__(key, value, step, **kwargs)


@attrs.define
class LogParam(Plugin):
    @override
    def __call__(self, key: str, value: Any, **kwargs) -> None:
        return super().__call__(key, value, **kwargs)


@attrs.define
class SetTag(Plugin):
    @override
    def __call__(self, key: str, value: Any, **kwargs) -> None:
        return super().__call__(key, value, **kwargs)


@attrs.define
class Start(Plugin):
    @override
    def __call__(self) -> None:
        return super().__call__()



from pathlib import Path
from typing import Any, override

import attrs
import mlflow
import mlflow.entities

from liblaf.cherries import info as _info
from liblaf.cherries import pathutils as _path
from liblaf.cherries.typed import PathLike

from ._abc import (
    End,
    LogArtifact,
    LogArtifacts,
    LogMetric,
    LogParam,
    RunStatus,
    SetTag,
    Start,
)


@attrs.define
class MlflowEnd(End):
    @override
    def __call__(self, status: RunStatus = RunStatus.FINISHED) -> None:
        mlflow.end_run(status)


@attrs.define
class MlflowLogArtifact(LogArtifact):
    @override
    def __call__(
        self, local_path: PathLike, artifact_path: PathLike | None = None, **kwargs
    ) -> Path:
        mlflow.log_artifact(
            _path.as_os_path(local_path), _path.as_posix(artifact_path), **kwargs
        )
        return _path.as_path(local_path)


@attrs.define
class MlflowLogArtifacts(LogArtifacts):
    @override
    def __call__(
        self, local_dir: PathLike, artifact_path: PathLike | None = None, **kwargs
    ) -> Path:
        mlflow.log_artifact(
            _path.as_os_path(local_dir), _path.as_posix(artifact_path), **kwargs
        )
        return _path.as_path(local_dir)


@attrs.define
class MlflowLogParam(LogParam):
    @override
    def __call__(self, key: str, value: Any, **kwargs) -> None:
        mlflow.log_param(key, value, **kwargs)


@attrs.define
class MlflowLogMetric(LogMetric):
    @override
    def __call__(
        self, key: str, value: float, step: int | None = None, **kwargs
    ) -> None:
        mlflow.log_metric(key, value, step, **kwargs)


@attrs.define
class MlflowSetTag(SetTag):
    @override
    def __call__(self, key: str, value: Any, **kwargs) -> None:
        mlflow.set_tag(key, value, **kwargs)


@attrs.define
class MlflowStart(Start):
    @override
    def __call__(self) -> None:
        mlflow.set_experiment(_info.exp_name())
        mlflow.start_run()



import subprocess as sp

import git

from liblaf import grapes
from liblaf.cherries import pathutils as _path


def git_auto_commit(
    message: str = "chore(cherries): auto commit", *, dry_run: bool = False
) -> None:
    repo: git.Repo = _repo()
    if not repo.is_dirty(untracked_files=True):
        return
    repo.git.add(all=True, dry_run=dry_run)
    sp.run(["git", "status"], check=False)
    if dry_run:
        return
    repo.git.commit(message=message)


def git_branch() -> str:
    repo: git.Repo = _repo()
    return repo.active_branch.name


def git_commit_sha() -> str:
    repo: git.Repo = _repo()
    return repo.head.commit.hexsha


def git_commit_url(sha: str | None = None) -> str:
    if sha is None:
        sha = git_commit_sha()
    info: grapes.git.GitInfo = git_info()
    if info.platform == "github":
        return f"https://github.com/{info.owner}/{info.repo}/commit/{sha}"
    raise NotImplementedError


def git_info() -> grapes.git.GitInfo:
    info: grapes.git.GitInfo = grapes.git.info(
        _path.entrypoint(absolute=True), search_parent_directories=True
    )
    return info


def _repo() -> git.Repo:
    return git.Repo(_path.entrypoint(absolute=True), search_parent_directories=True)



from ._exp_name import exp_name
from ._git import git_auto_commit, git_branch, git_commit_sha, git_commit_url, git_info

__all__ = [
    "exp_name",
    "git_auto_commit",
    "git_branch",
    "git_commit_sha",
    "git_commit_url",
    "git_info",
]



import git.exc
from environs import env

from liblaf import grapes

from ._git import git_info


def exp_name() -> str:
    if name := env.str("LIBLAF_CHERRIES_EXPERIMENT_NAME", "").strip():
        return name
    if name := env.str("MLFLOW_EXPERIMENT_NAME", "").strip():
        return name
    try:
        info: grapes.git.GitInfo = git_info()
    except git.exc.InvalidGitRepositoryError:
        return "Default"
    else:
        return info.repo



from typing import override

import attrs
from environs import env

from liblaf.cherries import info as _info

from ._abc import End, RunStatus, Start
from ._run import run


@attrs.define
class GitEnd(End):
    dry_run: bool = env.bool("LIBLAF_CHERRIES_GIT_DRY_RUN", default=False)

    @override
    def __call__(self, status: RunStatus = RunStatus.FINISHED) -> None:
        git_auto_commit(
            "chore(cherries): auto commit (on run end)", dry_run=self.dry_run
        )


@attrs.define
class GitStart(Start):
    dry_run: bool = env.bool("LIBLAF_CHERRIES_GIT_DRY_RUN", default=False)

    @override
    def __call__(self) -> None:
        git_auto_commit(
            "chore(cherries): auto commit (on run start)", dry_run=self.dry_run
        )


def git_auto_commit(
    header: str = "chore(cherries): auto commit", *, dry_run: bool = False
) -> None:
    body: str = ""
    if run.run_name and run.run_url:
        body += f"🏃 View run {run.run_name} at: {run.run_url}\n"
    if run.exp_name and run.exp_url:
        body += f"🧪 View experiment {run.exp_name} at: {run.exp_url}\n"
    message: str = f"{header}\n\n{body}" if body else header
    _info.git_auto_commit(message, dry_run=dry_run)
    run.set_tag("cherries.git.branch", _info.git_branch())
    run.set_tag("cherries.git.sha", _info.git_commit_sha())
    run.set_tag("cherries.git.url", _info.git_commit_url())



from typing import override

import attrs
from loguru import logger

from liblaf import grapes

from ._abc import End, RunStatus, Start
from ._run import run


@attrs.define
class LoggingEnd(End):
    @override
    def __call__(self, status: RunStatus = RunStatus.FINISHED) -> None:
        logger.complete()
        run.log_artifact(run.exp_dir / "run.log")
        run.log_artifact(run.exp_dir / "run.log.jsonl")


@attrs.define
class LoggingStart(Start):
    @override
    def __call__(self) -> None:
        grapes.init_logging(
            handlers=[
                grapes.logging.rich_handler(),
                grapes.logging.file_handler(file=run.exp_dir / "run.log"),
                grapes.logging.jsonl_handler(run.exp_dir / "run.log.jsonl"),
            ]
        )



from pathlib import Path

from liblaf.grapes.typed import PathLike

from ._path import exp_dir


def config(
    path: PathLike = "", *, mkdir: bool = True, prefix: PathLike = "config"
) -> Path:
    return _path(path, mkdir=mkdir, prefix=prefix)


def data(path: PathLike = "", *, mkdir: bool = True, prefix: PathLike = "data") -> Path:
    return _path(path, mkdir=mkdir, prefix=prefix)


def inputs(
    path: PathLike = "", *, mkdir: bool = True, prefix: PathLike = "data/inputs"
) -> Path:
    return _path(path, mkdir=mkdir, prefix=prefix)


def outputs(
    path: PathLike = "", *, mkdir: bool = True, prefix: PathLike = "data/outputs"
) -> Path:
    return _path(path, mkdir=mkdir, prefix=prefix)


def params(
    path: PathLike = "", *, mkdir: bool = True, prefix: PathLike = "params"
) -> Path:
    return _path(path, mkdir=mkdir, prefix=prefix)


def path(path: PathLike = "", *, mkdir: bool = True, prefix: PathLike = "") -> Path:
    return _path(path, mkdir=mkdir, prefix=prefix)


def src(path: PathLike = "", *, mkdir: bool = True, prefix: PathLike = "src") -> Path:
    return _path(path, mkdir=mkdir, prefix=prefix)


def _path(path: PathLike = "", *, mkdir: bool = True, prefix: PathLike = "") -> Path:
    path = Path(path)
    if not path.is_absolute():
        path = exp_dir(absolute=True) / prefix / path
    if mkdir:
        path.parent.mkdir(parents=True, exist_ok=True)
    return path



from ._abc import (
    End,
    LogArtifact,
    LogArtifacts,
    LogMetric,
    LogParam,
    Plugin,
    RunStatus,
    Start,
)
from ._dvc import DvcEnd, DvcLogArtifact, DvcLogArtifacts
from ._git import GitEnd, GitStart
from ._logging import LoggingEnd, LoggingStart
from ._mlflow import (
    MlflowEnd,
    MlflowLogArtifact,
    MlflowLogArtifacts,
    MlflowLogMetric,
    MlflowLogParam,
    MlflowSetTag,
    MlflowStart,
)
from ._run import (
    Run,
    end,
    log_artifact,
    log_artifacts,
    log_input,
    log_inputs,
    log_metric,
    log_output,
    log_outputs,
    log_param,
    log_src,
    run,
    set_tag,
    start,
)

__all__ = [
    "DvcEnd",
    "DvcLogArtifact",
    "DvcLogArtifacts",
    "End",
    "GitEnd",
    "GitStart",
    "LogArtifact",
    "LogArtifacts",
    "LogMetric",
    "LogParam",
    "LoggingEnd",
    "LoggingStart",
    "MlflowEnd",
    "MlflowLogArtifact",
    "MlflowLogArtifacts",
    "MlflowLogMetric",
    "MlflowLogParam",
    "MlflowSetTag",
    "MlflowStart",
    "Plugin",
    "Run",
    "RunStatus",
    "Start",
    "end",
    "log_artifact",
    "log_artifacts",
    "log_input",
    "log_inputs",
    "log_metric",
    "log_output",
    "log_outputs",
    "log_param",
    "log_src",
    "run",
    "set_tag",
    "start",
]
